import logging
import typing as t

__all__: t.Tuple[str, ...] = ("create_logging_setup",)


class Logger(logging.Formatter):
    """
    Logging formatter for Pokelance


    Attributes
    ----------
    COLOR_CONFIGS : t.Dict[int, str]
        A dictionary of logging levels and their respective color codes

    Examples
    -----------

    Color codes are used to color the log messages in the console.

    >>> COLOR_CONFIGS: t.Dict[int, str] = {
    ...     logging.DEBUG: "\x1b[36;49m",  # cyan
    ...     logging.INFO: "\x1b[32;49m",  # green
    ...     logging.WARNING: "\x1b[33;49m",  # yellow
    ...     logging.ERROR: "\x1b[31;49m",  # red
    ...     logging.CRITICAL: "\x1b[31;49m",  # red
    ... }
    """

    COLOR_CONFIGS: t.Dict[int, str] = {
        logging.DEBUG: "\x1b[33;49m",
        logging.INFO: "\x1b[32;49m",
        logging.WARNING: "\x1b[35;49m",
        logging.ERROR: "\x1b[31;49m",
        logging.CRITICAL: "\x1b[33;41;1m",
    }

    def format(self, record: logging.LogRecord) -> str:
        log_format = f"[%(asctime)s : {record.levelname.rjust(7)}] | %(message)s "

        formatter = logging.Formatter(
            "".join(
                (
                    self.COLOR_CONFIGS.get(record.levelno, "\x1b[32;49m"),
                    log_format,
                    "\x1b[0m",
                )
            )
        )
        return formatter.format(record)


def create_logging_setup(name: str) -> logging.Logger:
    """
    Create a logging setup for Pokelance

    Parameters
    ----------
    name: str
        The name of the logger

    Returns
    -------
    logging.Logger
        The logger

    Examples
    --------
    >>> logger = create_logging_setup("test")
    >>> logger.info("test")
    [2020-01-01 00:00:00 :  INFO] | test

    """
    _LOGGER = logging.getLogger(name)
    stream = logging.StreamHandler()
    stream.setFormatter(Logger())
    _LOGGER.addHandler(stream)
    _LOGGER.setLevel(logging.DEBUG)
    return _LOGGER



import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import (
    Description,
    Effect,
    GenerationGameIndex,
    MachineVersionDetail,
    Name,
    NamedResource,
    VerboseEffect,
    VersionGroupFlavorText,
)

from .utils import ItemHolderPokemon, ItemSprites

__all__: t.Tuple[str, ...] = ("Item", "ItemAttribute", "ItemCategory", "ItemFlingEffect", "ItemPocket")


@attrs.define(slots=True, kw_only=True)
class Item(BaseModel):
    """Item model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    cost: int
        The price of this item in stores.
    fling_power: int
        The power of the move Fling when used with this item.
    fling_effect: NamedResource
        The effect of the move Fling when used with this item.
    attributes: t.List[NamedResource]
        A list of attributes this item has.
    category: NamedResource
        The category of items this item falls into.
    effect_entries: t.List[Effect]
        The effect of this ability listed in different languages.
    flavor_text_entries: t.List[VersionGroupFlavorText]
        The flavor text of this ability listed in different languages.
    game_indices: t.List[GenerationGameIndex]
        A list of game indices relevent to this item by generation.
    names: t.List[Name]
        The name of this resource listed in different languages.
    sprites: ItemSprites
        A set of sprites used to depict this item in the game.
    held_by_pokemon: t.List[ItemHolderPokemon]
        A list of Pokémon that might be found in the wild holding this item.
    baby_trigger_for: NamedResource
        An evolution chain this item requires to produce a bay during mating.
    machines: t.List[MachineVersionDetail]
        A list of the machines related to this item.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    cost: int = attrs.field(factory=int)
    fling_power: int = attrs.field(factory=int)
    fling_effect: NamedResource = attrs.field(factory=NamedResource)
    attributes: t.List[NamedResource] = attrs.field(factory=list)
    category: NamedResource = attrs.field(factory=NamedResource)
    effect_entries: t.List[VerboseEffect] = attrs.field(factory=list)
    flavor_text_entries: t.List[VersionGroupFlavorText] = attrs.field(factory=list)
    game_indices: t.List[GenerationGameIndex] = attrs.field(factory=list)
    names: t.List[Name] = attrs.field(factory=list)
    sprites: ItemSprites = attrs.field(factory=ItemSprites)
    held_by_pokemon: t.List[ItemHolderPokemon] = attrs.field(factory=list)
    baby_trigger_for: NamedResource = attrs.field(factory=NamedResource)
    machines: t.List[MachineVersionDetail] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Item":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            cost=payload.get("cost", 0),
            fling_power=payload.get("fling_power", 0),
            fling_effect=NamedResource.from_payload(payload.get("fling_effect", {}) or {}),
            attributes=[NamedResource.from_payload(attribute) for attribute in payload.get("attributes", [])],
            category=NamedResource.from_payload(payload.get("category", {}) or {}),
            effect_entries=[VerboseEffect.from_payload(effect) for effect in payload.get("effect_entries", [])],
            flavor_text_entries=[
                VersionGroupFlavorText.from_payload(flavor_text)
                for flavor_text in payload.get("flavor_text_entries", [])
            ],
            game_indices=[
                GenerationGameIndex.from_payload(game_index) for game_index in payload.get("game_indices", [])
            ],
            names=[Name.from_payload(name) for name in payload.get("names", [])],
            sprites=ItemSprites.from_payload(payload.get("sprites", {})),
            held_by_pokemon=[
                ItemHolderPokemon.from_payload(held_by_pokemon)
                for held_by_pokemon in payload.get("held_by_pokemon", [])
            ],
            baby_trigger_for=NamedResource.from_payload(payload.get("baby_trigger_for", {}) or {}),
            machines=[MachineVersionDetail.from_payload(machine) for machine in payload.get("machines", [])],
        )


@attrs.define(slots=True, kw_only=True)
class ItemAttribute(BaseModel):
    """ItemAttribute model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    items: t.List[NamedResource]
        A list of items that have this attribute.
    names: t.List[Name]
        The name of this resource listed in different languages.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    items: t.List[NamedResource] = attrs.field(factory=list)
    names: t.List[Name] = attrs.field(factory=list)
    descriptions: t.List[Description] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "ItemAttribute":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            items=[NamedResource.from_payload(item) for item in payload.get("items", [])],
            names=[Name.from_payload(name) for name in payload.get("names", [])],
            descriptions=[Description.from_payload(description) for description in payload.get("descriptions", [])],
        )


@attrs.define(slots=True, kw_only=True)
class ItemCategory(BaseModel):
    """ItemCategory model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    items: t.List[NamedResource]
        A list of items that are a part of this category.
    names: t.List[Name]
        The name of this resource listed in different languages.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    items: t.List[NamedResource] = attrs.field(factory=list)
    names: t.List[Name] = attrs.field(factory=list)
    pocket: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "ItemCategory":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            items=[NamedResource.from_payload(item) for item in payload.get("items", [])],
            names=[Name.from_payload(name) for name in payload.get("names", [])],
            pocket=NamedResource.from_payload(payload.get("pocket", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class ItemFlingEffect(BaseModel):
    """ItemFlingEffect model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    effect_entries: t.List[Effect]
        The result of this fling effect listed in different languages.
    items: t.List[NamedResource]
        A list of items that have this fling effect.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    effect_entries: t.List[Effect] = attrs.field(factory=list)
    items: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "ItemFlingEffect":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            effect_entries=[Effect.from_payload(effect) for effect in payload.get("effect_entries", [])],
            items=[NamedResource.from_payload(item) for item in payload.get("items", [])],
        )


@attrs.define(slots=True, kw_only=True)
class ItemPocket(BaseModel):
    """ItemPocket model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    categories: t.List[NamedResource]
        A list of item categories that are relevant to this item pocket.
    names: t.List[Name]
        The name of this resource listed in different languages.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    categories: t.List[NamedResource] = attrs.field(factory=list)
    names: t.List[Name] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "ItemPocket":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            categories=[NamedResource.from_payload(category) for category in payload.get("categories", [])],
            names=[Name.from_payload(name) for name in payload.get("names", [])],
        )



import typing as t

from pokelance.http import Endpoint
from pokelance.models import EncounterCondition, EncounterConditionValue, EncounterMethod

from ._base import BaseExtension

if t.TYPE_CHECKING:
    from pokelance import PokeLance
    from pokelance.http import HttpClient


__all__: t.Tuple[str, ...] = (
    "setup",
    "Encounter",
)


class Encounter(BaseExtension):
    def __init__(self, client: "HttpClient") -> None:
        super().__init__(client)
        self.cache = self._cache.encounter

    async def setup(self) -> None:
        data = await self._client.request(Endpoint.get_encounter_condition_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "encounter_condition", data["results"])
        data = await self._client.request(Endpoint.get_encounter_method_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "encounter_method", data["results"])
        data = await self._client.request(Endpoint.get_encounter_condition_value_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "encounter_condition_value", data["results"])

    def get_encounter_condition(self, name: t.Union[str, int]) -> t.Optional[EncounterCondition]:
        """Gets an encounter condition from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the encounter condition.

        Returns
        -------
        typing.Optional[pokelance.models.EncounterCondition]
            The encounter condition if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the encounter condition is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> condition = client.encounter.get_encounter_condition("swarm")
        >>> condition.name
        'swarm'
        """
        route = Endpoint.get_encounter_condition(name)
        self._validate_resource(self.cache.encounter_condition, name, route)
        return self.cache.encounter_condition.get(route, None)

    async def fetch_encounter_condition(self, name: t.Union[str, int]) -> t.Optional[EncounterCondition]:
        """Fetches an encounter condition from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the encounter condition.

        Returns
        -------
        typing.Optional[pokelance.models.EncounterCondition]
            The encounter condition if it exists in the API, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the encounter condition is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     condition = await client.encounter.fetch_encounter_condition("swarm")
        ...     print(condition.name)
        ...     await client.close()
        >>> asyncio.run(main())
        swarm
        """
        route = Endpoint.get_encounter_condition(name)
        self._validate_resource(self.cache.encounter_condition, name, route)
        data = await self._client.request(route)
        return self.cache.encounter_condition.setdefault(route, EncounterCondition.from_payload(data))

    def get_encounter_condition_value(self, name: t.Union[str, int]) -> t.Optional[EncounterConditionValue]:
        """Gets an encounter condition value from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the encounter condition value.

        Returns
        -------
        typing.Optional[pokelance.models.EncounterConditionValue]
            The encounter condition value if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the encounter condition value is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> condition = client.encounter.get_encounter_condition_value("swarm-yes")
        >>> condition.name
        'swarm-yes'
        """
        route = Endpoint.get_encounter_condition_value(name)
        self._validate_resource(self.cache.encounter_condition_value, name, route)
        return self.cache.encounter_condition_value.get(route, None)

    async def fetch_encounter_condition_value(self, name: t.Union[str, int]) -> t.Optional[EncounterConditionValue]:
        """Fetches an encounter condition value from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the encounter condition value.

        Returns
        -------
        typing.Optional[pokelance.models.EncounterConditionValue]
            The encounter condition value if it exists in the API, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the encounter condition value is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     condition = await client.encounter.fetch_encounter_condition_value("swarm-yes")
        ...     print(condition.name)
        ...     await client.close()
        >>> asyncio.run(main())
        swarm-yes
        """
        route = Endpoint.get_encounter_condition_value(name)
        self._validate_resource(self.cache.encounter_condition_value, name, route)
        data = await self._client.request(route)
        return self.cache.encounter_condition_value.setdefault(route, EncounterConditionValue.from_payload(data))

    def get_encounter_method(self, name: t.Union[str, int]) -> t.Optional[EncounterMethod]:
        """Gets an encounter method from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the encounter method.

        Returns
        -------
        typing.Optional[pokelance.models.EncounterMethod]
            The encounter method if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the encounter method is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> method = client.encounter.get_encounter_method("walk")
        >>> method.name
        'walk'
        """
        route = Endpoint.get_encounter_method(name)
        self._validate_resource(self.cache.encounter_method, name, route)
        return self.cache.encounter_method.get(route, None)

    async def fetch_encounter_method(self, name: t.Union[str, int]) -> t.Optional[EncounterMethod]:
        """Fetches an encounter method from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the encounter method.

        Returns
        -------
        typing.Optional[pokelance.models.EncounterMethod]
            The encounter method if it exists in the API, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the encounter method is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     method = await client.encounter.fetch_encounter_method("walk")
        ...     print(method.name)
        ...     await client.close()
        >>> asyncio.run(main())
        walk
        """
        route = Endpoint.get_encounter_method(name)
        self._validate_resource(self.cache.encounter_method, name, route)
        data = await self._client.request(route)
        return self.cache.encounter_method.setdefault(route, EncounterMethod.from_payload(data))


def setup(lance: "PokeLance") -> None:
    """Sets up the encounter cog."""
    lance.add_extension("encounter", Encounter(lance.http))



import typing as t

from .models import (
    Description,
    Effect,
    Encounter,
    FlavorText,
    GenerationGameIndex,
    MachineVersionDetail,
    Name,
    VerboseEffect,
    VersionEncounterDetail,
    VersionGameIndex,
    VersionGroupFlavorText,
)
from .resources import NamedResource, Resource

__all__: t.Tuple[str, ...] = (
    "Description",
    "Effect",
    "Encounter",
    "FlavorText",
    "GenerationGameIndex",
    "MachineVersionDetail",
    "Name",
    "VerboseEffect",
    "VersionEncounterDetail",
    "VersionGameIndex",
    "VersionGroupFlavorText",
    "Resource",
    "NamedResource",
)



import typing as t

from pokelance.http import Endpoint
from pokelance.models import Move as MoveModel
from pokelance.models import MoveAilment, MoveBattleStyle, MoveCategory, MoveDamageClass, MoveLearnMethod, MoveTarget

from ._base import BaseExtension

if t.TYPE_CHECKING:
    from pokelance import PokeLance
    from pokelance.http import HttpClient


__all__: t.Tuple[str, ...] = (
    "setup",
    "Move",
)


class Move(BaseExtension):
    """Extension for move related endpoints.

    Parameters
    ----------
    client: pokelance.http.HttpClient
        The client to use for requests.

    Attributes
    ----------
    cache: pokelance.cache.Move
        The cache for this extension.
    """

    def __init__(self, client: "HttpClient") -> None:
        """Initializes the extension."""
        super().__init__(client)
        self.cache = self._cache.move

    async def setup(self) -> None:
        """Sets up the extension."""
        data = await self._client.request(Endpoint.get_move_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "move", data["results"])
        data = await self._client.request(Endpoint.get_move_ailment_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "move_ailment", data["results"])
        data = await self._client.request(Endpoint.get_move_battle_style_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "move_battle_style", data["results"])
        data = await self._client.request(Endpoint.get_move_category_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "move_category", data["results"])
        data = await self._client.request(Endpoint.get_move_damage_class_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "move_damage_class", data["results"])
        data = await self._client.request(Endpoint.get_move_learn_method_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "move_learn_method", data["results"])
        data = await self._client.request(Endpoint.get_move_target_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "move_target", data["results"])

    def get_move(self, name: t.Union[str, int]) -> t.Optional[MoveModel]:
        """Gets a move from the cache.

        Parameters
        ----------
        name: t.Union[str, int]
            The id of the move.

        Returns
        -------
        typing.Optional[pokelance.models.Move]
            The move if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the move does not exist in the cache.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> move = client.move.get_move(1)
        >>> move.name
        'pound'
        """
        route = Endpoint.get_move(name)
        self._validate_resource(self.cache.move, name, route)
        return self.cache.move.get(route, None)

    async def fetch_move(self, name: t.Union[str, int]) -> t.Optional[MoveModel]:
        """Fetches a move from the API.

        Parameters
        ----------
        name: t.Union[str, int]
            The id of the move.

        Returns
        -------
        typing.Optional[pokelance.models.Move]
            The move if it exists in the API, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the move does not exist in the API.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     move = await client.move.fetch_move(1)
        ...     print(move.name)
        ...     await client.close()
        >>> asyncio.run(main())
        pound
        """
        route = Endpoint.get_move(name)
        self._validate_resource(self.cache.move, name, route)
        data = await self._client.request(route)
        return self.cache.move.setdefault(route, MoveModel.from_payload(data))

    def get_move_ailment(self, name: t.Union[str, int]) -> t.Optional[MoveAilment]:
        """Gets a move ailment from the cache.

        Parameters
        ----------
        name: t.Union[str, int]
            The id of the move ailment.

        Returns
        -------
        typing.Optional[pokelance.models.MoveAilment]
            The move ailment if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the move ailment does not exist in the cache.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> ailment = client.move.get_move_ailment(1)
        >>> ailment.name
        'paralysis'
        """
        route = Endpoint.get_move_ailment(name)
        self._validate_resource(self.cache.move_ailment, name, route)
        return self.cache.move_ailment.get(route, None)

    async def fetch_move_ailment(self, name: t.Union[str, int]) -> t.Optional[MoveAilment]:
        """Fetches a move ailment from the API.

        Parameters
        ----------
        name: t.Union[str, int]
            The id of the move ailment.

        Returns
        -------
        typing.Optional[pokelance.models.MoveAilment]
            The move ailment if it exists in the API, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the move ailment does not exist in the API.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     ailment = await client.move.fetch_move_ailment(1)
        ...     print(ailment.name)
        ...     await client.close()
        >>> asyncio.run(main())
        paralysis
        """
        route = Endpoint.get_move_ailment(name)
        self._validate_resource(self.cache.move_ailment, name, route)
        data = await self._client.request(route)
        return self.cache.move_ailment.setdefault(route, MoveAilment.from_payload(data))

    def get_move_battle_style(self, name: t.Union[str, int]) -> t.Optional[MoveBattleStyle]:
        """Gets a move battle style from the cache.

        Parameters
        ----------
        name: t.Union[str, int]
            The id of the move battle style.

        Returns
        -------
        typing.Optional[pokelance.models.MoveBattleStyle]
            The move battle style if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the move battle style does not exist in the cache.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> style = client.move.get_move_battle_style(1)
        >>> style.name
        'attack'
        """
        route = Endpoint.get_move_battle_style(name)
        self._validate_resource(self.cache.move_battle_style, name, route)
        return self.cache.move_battle_style.get(route, None)

    async def fetch_move_battle_style(self, name: t.Union[str, int]) -> t.Optional[MoveBattleStyle]:
        """Fetches a move battle style from the API.

        Parameters
        ----------
        name: t.Union[str, int]
            The id of the move battle style.

        Returns
        -------
        typing.Optional[pokelance.models.MoveBattleStyle]
            The move battle style if it exists in the API, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the move battle style does not exist in the API.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     style = await client.move.fetch_move_battle_style(1)
        ...     print(style.name)
        ...     await client.close()
        >>> asyncio.run(main())
        attack
        """
        route = Endpoint.get_move_battle_style(name)
        self._validate_resource(self.cache.move_battle_style, name, route)
        data = await self._client.request(route)
        return self.cache.move_battle_style.setdefault(route, MoveBattleStyle.from_payload(data))

    def get_move_category(self, name: t.Union[str, int]) -> t.Optional[MoveCategory]:
        """Gets a move category from the cache.

        Parameters
        ----------
        name: t.Union[str, int]
            The id of the move category.

        Returns
        -------
        typing.Optional[pokelance.models.MoveCategory]
            The move category if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the move category does not exist in the cache.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> category = client.move.get_move_category(1)
        >>> category.name
        'ailment'
        """
        route = Endpoint.get_move_category(name)
        self._validate_resource(self.cache.move_category, name, route)
        return self.cache.move_category.get(route, None)

    async def fetch_move_category(self, name: t.Union[str, int]) -> t.Optional[MoveCategory]:
        """Fetches a move category from the API.

        Parameters
        ----------
        name: t.Union[str, int]
            The id of the move category.

        Returns
        -------
        typing.Optional[pokelance.models.MoveCategory]
            The move category if it exists in the API, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the move category does not exist in the API.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     category = await client.move.fetch_move_category(1)
        ...     print(category.name)
        ...     await client.close()
        >>> asyncio.run(main())
        ailment
        """
        route = Endpoint.get_move_category(name)
        self._validate_resource(self.cache.move_category, name, route)
        data = await self._client.request(route)
        return self.cache.move_category.setdefault(route, MoveCategory.from_payload(data))

    def get_move_damage_class(self, name: t.Union[str, int]) -> t.Optional[MoveDamageClass]:
        """Gets a move damage class from the cache.

        Parameters
        ----------
        name: t.Union[str, int]
            The id of the move damage class.

        Returns
        -------
        typing.Optional[pokelance.models.MoveDamageClass]
            The move damage class if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the move damage class does not exist in the cache.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> damage_class = client.move.get_move_damage_class(1)
        >>> damage_class.name
        'status'
        """
        route = Endpoint.get_move_damage_class(name)
        self._validate_resource(self.cache.move_damage_class, name, route)
        return self.cache.move_damage_class.get(route, None)

    async def fetch_move_damage_class(self, name: t.Union[str, int]) -> t.Optional[MoveDamageClass]:
        """Fetches a move damage class from the API.

        Parameters
        ----------
        name: t.Union[str, int]
            The id of the move damage class.

        Returns
        -------
        typing.Optional[pokelance.models.MoveDamageClass]
            The move damage class if it exists in the API, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the move damage class does not exist in the API.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     damage_class = await client.move.fetch_move_damage_class(1)
        ...     print(damage_class.name)
        ...     await client.close()
        >>> asyncio.run(main())
        status
        """
        route = Endpoint.get_move_damage_class(name)
        self._validate_resource(self.cache.move_damage_class, name, route)
        data = await self._client.request(route)
        return self.cache.move_damage_class.setdefault(route, MoveDamageClass.from_payload(data))

    def get_move_learn_method(self, name: t.Union[str, int]) -> t.Optional[MoveLearnMethod]:
        """Gets a move learn method from the cache.

        Parameters
        ----------
        name: t.Union[str, int]
            The id of the move learn method.

        Returns
        -------
        typing.Optional[pokelance.models.MoveLearnMethod]
            The move learn method if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the move learn method does not exist in the cache.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> learn_method = client.move.get_move_learn_method(1)
        >>> learn_method.name
        'level-up'
        """
        route = Endpoint.get_move_learn_method(name)
        self._validate_resource(self.cache.move_learn_method, name, route)
        return self.cache.move_learn_method.get(route, None)

    async def fetch_move_learn_method(self, name: t.Union[str, int]) -> t.Optional[MoveLearnMethod]:
        """Fetches a move learn method from the API.

        Parameters
        ----------
        name: t.Union[str, int]
            The id of the move learn method.

        Returns
        -------
        typing.Optional[pokelance.models.MoveLearnMethod]
            The move learn method if it exists in the API, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the move learn method does not exist in the API.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     learn_method = await client.move.fetch_move_learn_method(1)
        ...     print(learn_method.name)
        ...     await client.close()
        >>> asyncio.run(main())
        level-up
        """
        route = Endpoint.get_move_learn_method(name)
        self._validate_resource(self.cache.move_learn_method, name, route)
        data = await self._client.request(route)
        return self.cache.move_learn_method.setdefault(route, MoveLearnMethod.from_payload(data))

    def get_move_target(self, name: t.Union[str, int]) -> t.Optional[MoveTarget]:
        """Gets a move target from the cache.

        Parameters
        ----------
        name: t.Union[str, int]
            The id of the move target.

        Returns
        -------
        typing.Optional[pokelance.models.MoveTarget]
            The move target if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the move target does not exist in the cache.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> target = client.move.get_move_target(1)
        >>> target.name
        'specific-pokemon'
        """
        route = Endpoint.get_move_target(name)
        self._validate_resource(self.cache.move_target, name, route)
        return self.cache.move_target.get(route, None)

    async def fetch_move_target(self, name: t.Union[str, int]) -> t.Optional[MoveTarget]:
        """Fetches a move target from the API.

        Parameters
        ----------
        name: t.Union[str, int]
            The id of the move target.

        Returns
        -------
        typing.Optional[pokelance.models.MoveTarget]
            The move target if it exists in the API, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            If the move target does not exist in the API.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     target = await client.move.fetch_move_target(1)
        ...     print(target.name)
        ...     await client.close()
        >>> asyncio.run(main())
        specific-pokemon
        """
        route = Endpoint.get_move_target(name)
        self._validate_resource(self.cache.move_target, name, route)
        data = await self._client.request(route)
        return self.cache.move_target.setdefault(route, MoveTarget.from_payload(data))


def setup(lance: "PokeLance") -> None:
    """Sets up the move cog."""
    lance.add_extension("move", Move(lance.http))



import typing as t

from ._base import BaseModel
from .abstract import (
    Ability,
    Berry,
    BerryFirmness,
    BerryFlavor,
    Characteristic,
    ContestEffect,
    ContestType,
    EggGroup,
    EncounterCondition,
    EncounterConditionValue,
    EncounterMethod,
    EvolutionChain,
    EvolutionTrigger,
    Gender,
    Generation,
    GrowthRate,
    Item,
    ItemAttribute,
    ItemCategory,
    ItemFlingEffect,
    ItemPocket,
    Location,
    LocationArea,
    LocationAreaEncounter,
    Machine,
    Move,
    MoveAilment,
    MoveBattleStyle,
    MoveCategory,
    MoveDamageClass,
    MoveLearnMethod,
    MoveTarget,
    Nature,
    PalParkArea,
    PokeathlonStat,
    Pokedex,
    Pokemon,
    PokemonColor,
    PokemonForm,
    PokemonHabitats,
    PokemonShape,
    PokemonSpecies,
    Region,
    Stat,
    SuperContestEffect,
    Type,
    Version,
    VersionGroup,
)

__all__: t.Tuple[str, ...] = (
    "BaseModel",
    "Berry",
    "BerryFirmness",
    "BerryFlavor",
    "ContestEffect",
    "ContestType",
    "SuperContestEffect",
    "EncounterCondition",
    "EncounterConditionValue",
    "EncounterMethod",
    "EvolutionChain",
    "EvolutionTrigger",
    "Generation",
    "Pokedex",
    "Version",
    "VersionGroup",
    "Item",
    "ItemAttribute",
    "ItemCategory",
    "ItemFlingEffect",
    "ItemPocket",
    "Location",
    "LocationArea",
    "PalParkArea",
    "Region",
    "Machine",
    "Move",
    "MoveAilment",
    "MoveBattleStyle",
    "MoveCategory",
    "MoveDamageClass",
    "MoveLearnMethod",
    "MoveTarget",
    "Ability",
    "Characteristic",
    "Pokemon",
    "PokemonSpecies",
    "PokemonForm",
    "Nature",
    "Type",
    "Gender",
    "GrowthRate",
    "EggGroup",
    "LocationAreaEncounter",
    "PokemonColor",
    "PokeathlonStat",
    "PokemonHabitats",
    "PokemonShape",
    "Stat",
)



import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import Name, NamedResource

from .utils import ChainLink

__all__: t.Tuple[str, ...] = (
    "EvolutionChain",
    "EvolutionTrigger",
)


@attrs.define(slots=True, kw_only=True)
class EvolutionChain(BaseModel):
    """Evolution chain model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    baby_trigger_item: t.Optional[NamedResource]
        The item that a Pokémon would be holding when mating that would trigger the egg hatching a baby
         Pokémon rather than a basic Pokémon.
    chain: ChainLink
        The base chain link object. Each link contains evolution details for a Pokémon in the chain.
         Each link references the next Pokémon in the natural evolution order.
    """

    id: int = attrs.field(factory=int)
    baby_trigger_item: NamedResource = attrs.field(factory=NamedResource)
    chain: ChainLink = attrs.field(factory=ChainLink)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "EvolutionChain":
        return cls(
            id=payload.get("id", 0),
            baby_trigger_item=NamedResource.from_payload(payload.get("baby_trigger_item", {}) or {}),
            chain=ChainLink.from_payload(payload.get("chain", {})),
        )


@attrs.define(slots=True, kw_only=True)
class EvolutionTrigger(BaseModel):
    """Evolution trigger model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    names: t.List[Name]
        A list of name and language pairs for this resource.
    pokemon_species: t.List[NamedResource]
        A list of pokemon species that result from this evolution trigger.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    names: t.List[Name] = attrs.field(factory=list)
    pokemon_species: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "EvolutionTrigger":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            names=[Name.from_payload(name) for name in payload.get("names", [])],
            pokemon_species=[NamedResource.from_payload(species) for species in payload.get("pokemon_species", [])],
        )



import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import GenerationGameIndex, Name, NamedResource

from .utils import EncounterMethodRate, PalParkEncounterSpecies, PokemonEncounter

__all__: t.Tuple[str, ...] = (
    "Location",
    "LocationArea",
    "PalParkArea",
    "Region",
)


@attrs.define(slots=True, kw_only=True)
class Location(BaseModel):
    """Location model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    region: NamedResource
        The region this location can be found in.
    names: t.List[Name]
        The name of this resource listed in different languages.
    game_indices: t.List[GenerationGameIndex]
        A list of game indices relevent to this location by generation.
    areas: t.List[NamedResource]
        A list of methods in which Pokémon may be encountered in this
        location.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    region: NamedResource = attrs.field(factory=NamedResource)
    names: t.List[Name] = attrs.field(factory=list)
    game_indices: t.List[GenerationGameIndex] = attrs.field(factory=list)
    areas: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Location":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            region=NamedResource.from_payload(payload.get("region", {}) or {}),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            game_indices=[GenerationGameIndex.from_payload(i) for i in payload.get("game_indices", [])],
            areas=[NamedResource.from_payload(i) for i in payload.get("areas", [])],
        )


@attrs.define(slots=True, kw_only=True)
class LocationArea(BaseModel):
    """LocationArea model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    game_index: int
        The internal id of an API resource within game data.
    encounter_method_rates: t.List[EncounterMethodRate]
        A list of methods in which Pokémon may be encountered in this
        area and how likely the method will occur depending on the
        version of the game.
    location: NamedResource
        The region this location can be found in.
    names: t.List[Name]
        The name of this resource listed in different languages.
    pokemon_encounters: t.List[PokemonEncounter]
        A list of Pokémon that can be encountered in this area along
        with version specific details about the encounter.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    game_index: int = attrs.field(factory=int)
    names: t.List[Name] = attrs.field(factory=list)
    location: NamedResource = attrs.field(factory=NamedResource)
    encounter_method_rates: t.List[EncounterMethodRate] = attrs.field(factory=list)
    pokemon_encounters: t.List[PokemonEncounter] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "LocationArea":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            game_index=payload.get("game_index", 0),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            location=NamedResource.from_payload(payload.get("location", {}) or {}),
            encounter_method_rates=[
                EncounterMethodRate.from_payload(i) for i in payload.get("encounter_method_rates", [])
            ],
            pokemon_encounters=[PokemonEncounter.from_payload(i) for i in payload.get("pokemon_encounters", [])],
        )


@attrs.define(slots=True, kw_only=True)
class PalParkArea(BaseModel):
    """PalParkArea model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    names: t.List[Name]
        The name of this resource listed in different languages.
    pokemon_encounters: t.List[PalParkEncounterSpecies]
        A list of Pokémon encountered in this pal park area along with
        details.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    names: t.List[Name] = attrs.field(factory=list)
    pokemon_encounters: t.List[PalParkEncounterSpecies] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PalParkArea":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            pokemon_encounters=[PalParkEncounterSpecies.from_payload(i) for i in payload.get("pokemon_encounters", [])],
        )


@attrs.define(slots=True, kw_only=True)
class Region(BaseModel):
    """Region model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    locations: t.List[NamedResource]
        A list of locations that can be found in this region.
    main_generation: NamedResource
        The generation this region was introduced in.
    names: t.List[Name]
        The name of this resource listed in different languages.
    pokedexes: t.List[NamedResource]
        A list of pokédexes that catalogue Pokémon in this region.
    version_groups: t.List[NamedResource]
        A list of version groups where this region can be visited.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    locations: t.List[NamedResource] = attrs.field(factory=list)
    main_generation: NamedResource = attrs.field(factory=NamedResource)
    names: t.List[Name] = attrs.field(factory=list)
    pokedexes: t.List[NamedResource] = attrs.field(factory=list)
    version_groups: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Region":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            locations=[NamedResource.from_payload(i) for i in payload.get("locations", [])],
            main_generation=NamedResource.from_payload(payload.get("main_generation", {}) or {}),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            pokedexes=[NamedResource.from_payload(i) for i in payload.get("pokedexes", [])],
            version_groups=[NamedResource.from_payload(i) for i in payload.get("version_groups", [])],
        )



import typing as t

import attrs


@attrs.define(hash=True, slots=True, kw_only=True)
class BaseModel:
    """Base model for all models"""

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "BaseModel":
        """Create a model from a payload

        Parameters
        ----------
        payload: typing.Dict[str, Any]
            The payload to create the model from.

        Returns
        -------
        BaseModel
            The model created from the payload.
        """
        return cls()



import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import NamedResource

__all__: t.Tuple[str, ...] = ("PokemonEntry",)


@attrs.define(slots=True, kw_only=True)
class PokemonEntry(BaseModel):
    """A pokemon entry resource.

    Attributes
    ----------
    entry_number: int
        The index of this Pokémon species entry within the Pokédex.
    pokemon_species: NamedResource
        The Pokémon species being encountered.
    """

    entry_number: int = attrs.field(factory=int)
    pokemon_species: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonEntry":
        return cls(
            entry_number=payload.get("entry_number", 0),
            pokemon_species=NamedResource.from_payload(payload.get("pokemon_species", {}) or {}),
        )



import json
import typing as t
from collections.abc import MutableMapping

import aiofiles

if t.TYPE_CHECKING:
    from pokelance import models  # noqa: F401
    from pokelance.http import Route  # noqa: F401


__all__: t.Tuple[str, ...] = (
    "PokemonCache",
    "BaseCache",
    "BerryCache",
    "PalParkAreaCache",
    "GenderCache",
    "GamesPokedexCache",
    "GamesVersionCache",
    "GamesGenerationCache",
    "BerryFlavorCache",
    "BerryFirmnessCache",
    "ContestTypeCache",
    "PokeathlonStatCache",
    "PokemonSpeciesCache",
    "PokemonLocationAreaCache",
    "ContestEffectCache",
    "SuperContestEffectCache",
    "PokemonColorCache",
    "PokemonFormCache",
    "PokemonHabitatCache",
    "PokemonShapeCache",
    "GrowthRateCache",
    "NatureCache",
    "TypeCache",
    "GenderCache",
    "GamesVersionGroupCache",
    "EggGroupCache",
    "EvolutionChainCache",
    "EvolutionTriggerCache",
    "EncounterConditionCache",
    "EncounterConditionValueCache",
    "EncounterMethodCache",
    "ItemCache",
    "ItemAttributeCache",
    "ItemCategoryCache",
    "ItemFlingEffectCache",
    "ItemPocketCache",
    "LocationCache",
    "LocationAreaCache",
    "MachineCache",
    "MoveCache",
    "MoveAilmentCache",
    "MoveBattleStyleCache",
    "MoveCategoryCache",
    "MoveDamageClassCache",
    "MoveLearnMethodCache",
    "MoveTargetCache",
    "AbilityCache",
    "CharacteristicCache",
    "PokemonCache",
    "PokemonSpeciesCache",
    "PokemonFormCache",
    "NatureCache",
    "TypeCache",
    "GenderCache",
    "GrowthRateCache",
    "PokemonLocationAreaCache",
    "StatCache",
    "RegionCache",
)

_KT = t.TypeVar("_KT")
_VT = t.TypeVar("_VT")


class BaseCache(MutableMapping[_KT, _VT]):
    """Base class for all caches.

    Parameters
    ----------
    max_size: int
        The maximum size of the cache.

    Attributes
    ----------
    _max_size: int
        The maximum size of the cache.
    _cache: typing.Dict[_KT, _VT]
        The cache itself.
    _endpoints: typing.Dict[str, int]
        The endpoints that are cached.
    """

    def __init__(self, max_size: int = 100) -> None:
        self._max_size = max_size
        self._cache: t.Dict[_KT, _VT] = {}
        self._endpoints: t.Dict[str, int] = {}

    def __getitem__(self, key: _KT) -> _VT:
        self._cache[key] = self._cache.pop(key)
        return self._cache[key]

    def __setitem__(self, key: _KT, value: _VT) -> None:
        if key in self._cache:
            self._cache[key] = self._cache.pop(key)
        else:
            if len(self._cache) >= self._max_size:
                self._cache.pop(list(self._cache.keys())[0])
            self._cache[key] = value

    def __delitem__(self, key: _KT) -> None:
        del self._cache[key]

    def __len__(self) -> int:
        return len(self._cache)

    def __iter__(self) -> t.Iterator[_KT]:
        return iter(self._cache)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._cache})"

    def setdefault(self, __key: _KT, __default: t.Any = ...) -> _VT:
        item = self._cache.setdefault(__key, __default)
        if item is __default:
            self._cache[__key] = self._cache.pop(__key)
        return item

    def load_documents(self, data: t.List[t.Dict[str, str]]) -> None:
        """Load documents into the cache.

        Parameters
        ----------
        data: typing.List[typing.Dict[str, str]]
            The data to load.
        """
        for document in data:
            self._endpoints[document["name"]] = int(document["url"].split("/")[-2])

    async def save(self, path: str = "") -> None:
        """Save the cache to a file.

        Parameters
        ----------
        path: str
            The path to save the cache to.
        """
        async with aiofiles.open(f"{path}{self.__class__.__name__}.json", "w") as f:
            await f.write(json.dumps(self._cache, indent=4, ensure_ascii=False))

    @property
    def endpoints(self) -> t.Dict[str, int]:
        """The endpoints that are cached.

        Returns
        -------
        typing.Dict[str, int]
            The endpoints that are cached.
        """
        return self._endpoints

    @property
    def cache(self) -> t.Dict[_KT, _VT]:
        """The cache itself.

        Returns
        -------
        typing.Dict[_KT, _VT]
            The cache itself.
        """
        return self._cache


class SecondaryTypeCache(BaseCache[_KT, _VT]):
    """A cache for secondary types with differing endpoints."""

    def load_documents(self, data: t.List[t.Dict[str, str]]) -> None:
        """Load documents into the cache. Endpoints are different for secondary types.

        Parameters
        ----------
        data: typing.List[typing.Dict[str, str]]
            The data to load.
        """
        for document in data:
            self._endpoints[document["url"].split("/")[-2]] = int(document["url"].split("/")[-2])


class BerryCache(BaseCache["Route", "models.Berry"]):
    """A cache for berries."""


class BerryFirmnessCache(BaseCache["Route", "models.BerryFirmness"]):
    """A cache for berry firmnesses."""


class BerryFlavorCache(BaseCache["Route", "models.BerryFlavor"]):
    """A cache for berry flavors."""


class ContestTypeCache(BaseCache["Route", "models.ContestType"]):
    """A cache for contest types."""


class PokemonCache(BaseCache["Route", "models.Pokemon"]):
    """A cache for pokemon."""


class AbilityCache(BaseCache["Route", "models.Ability"]):
    """A cache for abilities."""


class EggGroupCache(BaseCache["Route", "models.EggGroup"]):
    """A cache for egg groups."""


class GenderCache(BaseCache["Route", "models.Gender"]):
    """A cache for genders."""


class GrowthRateCache(BaseCache["Route", "models.GrowthRate"]):
    """A cache for growth rates."""


class NatureCache(BaseCache["Route", "models.Nature"]):
    """A cache for natures."""


class PokeathlonStatCache(BaseCache["Route", "models.PokeathlonStat"]):
    """A cache for pokeathlon stats."""


class PokemonColorCache(BaseCache["Route", "models.PokemonColor"]):
    """A cache for pokemon colors."""


class PokemonFormCache(BaseCache["Route", "models.PokemonForm"]):
    """A cache for pokemon forms."""


class PokemonLocationAreaCache(BaseCache["Route", "models.LocationAreaEncounter"]):
    """A cache for pokemon location areas."""


class PokemonHabitatCache(BaseCache["Route", "models.PokemonHabitats"]):
    """A cache for pokemon habitats."""


class PokemonShapeCache(BaseCache["Route", "models.PokemonShape"]):
    """A cache for pokemon shapes."""


class PokemonSpeciesCache(BaseCache["Route", "models.PokemonSpecies"]):
    """A cache for pokemon species."""


class StatCache(BaseCache["Route", "models.Stat"]):
    """A cache for stats."""


class TypeCache(BaseCache["Route", "models.Type"]):
    """A cache for types."""


class EncounterMethodCache(BaseCache["Route", "models.EncounterMethod"]):
    """A cache for encounter methods."""


class EncounterConditionCache(BaseCache["Route", "models.EncounterCondition"]):
    """A cache for encounter conditions."""


class EncounterConditionValueCache(BaseCache["Route", "models.EncounterConditionValue"]):
    """A cache for encounter condition values."""


class EvolutionTriggerCache(BaseCache["Route", "models.EvolutionTrigger"]):
    """A cache for evolution triggers."""


class GamesGenerationCache(BaseCache["Route", "models.Generation"]):
    """A cache for games generations."""


class GamesPokedexCache(BaseCache["Route", "models.Pokedex"]):
    """A cache for games pokedexes."""


class GamesVersionCache(BaseCache["Route", "models.Version"]):
    """A cache for games versions."""


class GamesVersionGroupCache(BaseCache["Route", "models.VersionGroup"]):
    """A cache for games version groups."""


class ItemCache(BaseCache["Route", "models.Item"]):
    """A cache for items."""


class ItemAttributeCache(BaseCache["Route", "models.ItemAttribute"]):
    """A cache for item attributes."""


class ItemCategoryCache(BaseCache["Route", "models.ItemCategory"]):
    """A cache for item categories."""


class ItemFlingEffectCache(BaseCache["Route", "models.ItemFlingEffect"]):
    """A cache for item fling effects."""


class ItemPocketCache(BaseCache["Route", "models.ItemPocket"]):
    """A cache for item pockets."""


class LocationCache(BaseCache["Route", "models.Location"]):
    """A cache for locations."""


class LocationAreaCache(BaseCache["Route", "models.LocationArea"]):
    """A cache for location areas."""


class PalParkAreaCache(BaseCache["Route", "models.PalParkArea"]):
    """A cache for pal park areas."""


class RegionCache(BaseCache["Route", "models.Region"]):
    """A cache for regions."""


class MoveCache(BaseCache["Route", "models.Move"]):
    """A cache for moves."""


class MoveAilmentCache(BaseCache["Route", "models.MoveAilment"]):
    """A cache for move ailments."""


class MoveBattleStyleCache(BaseCache["Route", "models.MoveBattleStyle"]):
    """A cache for move battle styles."""


class MoveCategoryCache(BaseCache["Route", "models.MoveCategory"]):
    """A cache for move categories."""


class MoveDamageClassCache(BaseCache["Route", "models.MoveDamageClass"]):
    """A cache for move damage classes."""


class MoveLearnMethodCache(BaseCache["Route", "models.MoveLearnMethod"]):
    """A cache for move learn methods."""


class MoveTargetCache(BaseCache["Route", "models.MoveTarget"]):
    """A cache for move targets."""


class MachineCache(SecondaryTypeCache["Route", "models.Machine"]):
    """A cache for machines."""


class EvolutionChainCache(SecondaryTypeCache["Route", "models.EvolutionChain"]):
    """A cache for evolution chains."""


class CharacteristicCache(SecondaryTypeCache["Route", "models.Characteristic"]):
    """A cache for characteristics."""


class ContestEffectCache(SecondaryTypeCache["Route", "models.ContestEffect"]):
    """A cache for contest effects."""


class SuperContestEffectCache(SecondaryTypeCache["Route", "models.SuperContestEffect"]):
    """A cache for super contest effects."""



import typing as t

from pokelance.http import Endpoint
from pokelance.models import Generation, Pokedex, Version, VersionGroup

from ._base import BaseExtension

if t.TYPE_CHECKING:
    from pokelance import PokeLance
    from pokelance.http import HttpClient


__all__: t.Tuple[str, ...] = (
    "setup",
    "Game",
)


class Game(BaseExtension):
    """Extension for game related endpoints.

    Parameters
    ----------
    client: pokelance.http.HttpClient
        The client to use for requests.

    Attributes
    ----------
    cache: pokelance.cache.Game
        The cache to use for caching resources.
    """

    def __init__(self, client: "HttpClient") -> None:
        """Initializes the game extension."""
        super().__init__(client)
        self.cache = self._cache.game

    async def setup(self) -> None:
        """Sets up the game extension."""
        data = await self._client.request(Endpoint.get_generation_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "generation", data["results"])
        data = await self._client.request(Endpoint.get_pokedex_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "pokedex", data["results"])
        data = await self._client.request(Endpoint.get_version_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "version", data["results"])
        data = await self._client.request(Endpoint.get_version_group_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "version_group", data["results"])

    def get_generation(self, name: t.Union[str, int]) -> t.Optional[Generation]:
        """Gets a generation from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the generation.

        Returns
        -------
        typing.Optional[pokelance.models.Generation]
            The generation if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the generation is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> generation = client.game.get_generation(1)
        >>> generation.id
        1
        """
        route = Endpoint.get_generation(name)
        self._validate_resource(self.cache.generation, name, route)
        return self.cache.generation.get(route, None)

    async def fetch_generation(self, name: t.Union[str, int]) -> t.Optional[Generation]:
        """Fetches a generation from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the generation.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the generation is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     generation = await client.game.fetch_generation(1)
        ...     print(generation.id)
        ...     await client.close()
        >>> asyncio.run(main())
        1
        """
        route = Endpoint.get_generation(name)
        self._validate_resource(self.cache.generation, name, route)
        data = await self._client.request(route)
        return self.cache.generation.setdefault(route, Generation.from_payload(data))

    def get_pokedex(self, name: t.Union[str, int]) -> t.Optional[Pokedex]:
        """Gets a pokedex from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the pokedex.

        Returns
        -------
        typing.Optional[pokelance.models.Pokedex]
            The pokedex if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the pokedex is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> pokedex = client.game.get_pokedex(1)
        >>> pokedex.region
        None
        """
        route = Endpoint.get_pokedex(name)
        self._validate_resource(self.cache.pokedex, name, route)
        return self.cache.pokedex.get(route, None)

    async def fetch_pokedex(self, name: t.Union[str, int]) -> t.Optional[Pokedex]:
        """Fetches a pokedex from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the pokedex.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the pokedex is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     pokedex = await client.game.fetch_pokedex(1)
        ...     print(pokedex.region)
        ...     await client.close()
        >>> asyncio.run(main())
        None
        """
        route = Endpoint.get_pokedex(name)
        self._validate_resource(self.cache.pokedex, name, route)
        data = await self._client.request(route)
        return self.cache.pokedex.setdefault(route, Pokedex.from_payload(data))

    def get_version(self, name: t.Union[str, int]) -> t.Optional[Version]:
        """Gets a version from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the version.

        Returns
        -------
        typing.Optional[pokelance.models.Version]
            The version if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the version is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> version = client.game.get_version(1)
        >>> version.name
        'red'
        """
        route = Endpoint.get_version(name)
        self._validate_resource(self.cache.version, name, route)
        return self.cache.version.get(route, None)

    async def fetch_version(self, name: t.Union[str, int]) -> t.Optional[Version]:
        """Fetches a version from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the version.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the version is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     version = await client.game.fetch_version(1)
        ...     print(version.name)
        ...     await client.close()
        >>> asyncio.run(main())
        red
        """
        route = Endpoint.get_version(name)
        self._validate_resource(self.cache.version, name, route)
        data = await self._client.request(route)
        return self.cache.version.setdefault(route, Version.from_payload(data))

    def get_version_group(self, name: t.Union[str, int]) -> t.Optional[VersionGroup]:
        """Gets a version group from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the version group.

        Returns
        -------
        typing.Optional[pokelance.models.VersionGroup]
            The version group if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the version group is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> version_group = client.game.get_version_group(1)
        >>> version_group.id
        1
        """
        route = Endpoint.get_version_group(name)
        self._validate_resource(self.cache.version_group, name, route)
        return self.cache.version_group.get(route, None)

    async def fetch_version_group(self, name: t.Union[str, int]) -> t.Optional[VersionGroup]:
        """Fetches a version group from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the version group.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the version group is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     version_group = await client.game.fetch_version_group(1)
        ...     print(version_group.id)
        ...     await client.close()
        >>> asyncio.run(main())
        1
        """
        route = Endpoint.get_version_group(name)
        self._validate_resource(self.cache.version_group, name, route)
        data = await self._client.request(route)
        return self.cache.version_group.setdefault(route, VersionGroup.from_payload(data))


def setup(lance: "PokeLance") -> None:
    """Sets up the game cog."""
    lance.add_extension("game", Game(lance.http))



import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import Name, NamedResource

__all__: t.Tuple[str, ...] = (
    "EncounterMethod",
    "EncounterCondition",
    "EncounterConditionValue",
)


@attrs.define(slots=True, kw_only=True)
class EncounterMethod(BaseModel):
    """An encounter method resource.

    Attributes
    ----------
    id: int
        The identifier for this encounter method resource.
    name: str
        The name for this encounter method resource.
    order: int
        A good value for sorting.
    names: typing.List[pokelance.models.common.Name]
        The name of this encounter method listed in different languages.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    order: int = attrs.field(factory=int)
    names: t.List[Name] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "EncounterMethod":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            order=payload.get("order", 0),
            names=[Name.from_payload(name) for name in payload.get("names", [])],
        )


@attrs.define(slots=True, kw_only=True)
class EncounterCondition(BaseModel):
    """An encounter condition resource.

    Attributes
    ----------
    id: int
        The identifier for this encounter condition resource.
    name: str
        The name for this encounter condition resource.
    names: typing.List[pokelance.models.common.Name]
        The name of this encounter condition listed in different languages.
    values: typing.List[pokelance.models.common.NamedResource]
        A list of possible values for this encounter condition.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    names: t.List[Name] = attrs.field(factory=list)
    values: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "EncounterCondition":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            names=[Name.from_payload(name) for name in payload.get("names", [])],
            values=[NamedResource.from_payload(value) for value in payload.get("values", [])],
        )


@attrs.define(slots=True, kw_only=True)
class EncounterConditionValue(BaseModel):
    """An encounter condition value resource.

    Attributes
    ----------
    id: int
        The identifier for this encounter condition value resource.
    name: str
        The name for this encounter condition value resource.
    condition: pokelance.models.common.NamedResource
        The condition this encounter condition value pertains to.
    names: typing.List[pokelance.models.common.Name]
        The name of this encounter condition value listed in different languages.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    condition: NamedResource = attrs.field(factory=NamedResource)
    names: t.List[Name] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "EncounterConditionValue":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            condition=NamedResource.from_payload(payload.get("condition", {}) or {}),
            names=[Name.from_payload(name) for name in payload.get("names", [])],
        )



import typing as t
from pathlib import Path

from .http import HttpClient
from .logger import create_logging_setup

if t.TYPE_CHECKING:
    import logging
    from types import TracebackType

    import aiohttp

    from .ext import BaseExtension, Berry, Contest, Encounter, Evolution, Game, Item, Location, Machine, Move, Pokemon

__all__: t.Tuple[str, ...] = ("PokeLance",)


class PokeLance:
    """
    Main class to interact with the PokeAPI.


    Attributes
    ----------
    _http : HttpClient
        The HTTP client used to make requests to the PokeAPI.
    _logger : logging.Logger
        The logger used to log information about the client.
    _loaders : t.List[t.Tuple[t.Coroutine[t.Any, t.Any, None], str]]
        A list of coroutines to load extension data.
    EXTENSIONS : Path
        The path to the extensions directory.
    berry : Berry
        The berry extension.
    contest : Contest
        The contest extension.
    encounter : Encounter
        The encounter extension.
    evolution : Evolution
        The evolution extension.
    game : Game
        The game extension.
    item : Item
        The item extension.
    location : Location
        The location extension.
    machine : Machine
        The machine extension.
    move : Move
        The move extension.
    pokemon : Pokemon
        The pokemon extension.

    Examples
    --------

    >>> import pokelance
    >>> import asyncio
    >>> client = pokelance.PokeLance()
    >>> async def main() -> None:
    ...     print(await client.ping())
    ...     await client.close()
    >>> asyncio.run(main())
    """

    EXTENSIONS: Path = Path(__file__).parent / "ext"
    berry: "Berry"
    contest: "Contest"
    evolution: "Evolution"
    game: "Game"
    item: "Item"
    location: "Location"
    machine: "Machine"
    move: "Move"
    pokemon: "Pokemon"
    encounter: "Encounter"

    def __init__(
        self,
        *,
        cache_size: int = 100,
        session: t.Optional["aiohttp.ClientSession"] = None,
    ) -> None:
        """
        Parameters
        ----------
        cache_size : int
            The size of the cache to use for the HTTP client.
        session : typing.Optional[aiohttp.ClientSession]
            The session to use for the HTTP client. It is recommended to use the default.

        Returns
        -------
        PokeLance
            The client.
        """
        self._logger = create_logging_setup("Client")
        self._http = HttpClient(client=self, session=session, cache_size=cache_size)
        self._loaders: t.List[t.Tuple[t.Coroutine[t.Any, t.Any, None], str]] = []
        self.setup_hook()

    async def __aenter__(self) -> "PokeLance":
        return self

    async def __aexit__(
        self,
        exc_type: t.Optional[t.Type[BaseException]],
        exc_val: t.Optional[BaseException],
        exc_tb: t.Optional["TracebackType"],
    ) -> None:
        if self._http.session is not None:
            await self._http.session.close()

    def setup_hook(self) -> None:
        """
        The setup hook to be called after the client is created.
        This is called automatically when the client is created.
        It is not recommended to call this manually.
        """
        self._logger.info(f"Using cache size: {self._http.cache.max_size}")
        for extension in self.EXTENSIONS.iterdir():
            if extension.is_file() and extension.suffix == ".py":
                if "_" not in extension.stem:
                    module = __import__(f"pokelance.ext.{extension.stem}", fromlist=["setup"])
                    module.setup(self)
        self._logger.info("Setup complete")

    def add_extension(self, name: str, extension: "BaseExtension") -> None:
        """
        Adds an extension to the client. This is called automatically when an extension is loaded.

        Parameters
        ----------
        name : str
            The name of the extension.
        extension : BaseExtension
            The extension to add.
        """
        self._loaders.append((extension.setup(), name))
        setattr(self, name, extension)

    async def ping(self) -> float:
        """
        Pings the PokeAPI and returns the latency.

        Returns
        -------
        float
            The latency of the PokeAPI.
        """
        return await self._http.ping()

    async def close(self) -> None:
        """
        Closes the client session. Recommended to use this when the client is no longer needed.
        Not needed if the client is used in a context manager.
        """
        self._logger.info("Closing session")
        if self._http.session is not None:
            await self._http.session.close()

    @property
    def loaders(self) -> t.List[t.Tuple[t.Coroutine[t.Any, t.Any, None], str]]:
        """
        The list of loaders for the extensions.

        Returns
        -------
        typing.List[typing.Tuple[typing.Coroutine[typing.Any, typing.Any, None], str]]
            The list of loaders.
        """
        return self._loaders

    @property
    def logger(self) -> "logging.Logger":
        """
        The logger used to log information about the client.

        Returns
        -------
        logging.Logger
            The logger.
        """
        return self._logger

    @property
    def http(self) -> HttpClient:
        """
        The HTTP client used to make requests to the PokeAPI.

        Returns
        -------
        HttpClient
            The HTTP client.
        """
        return self._http



import typing as t

import attrs

from pokelance.models import BaseModel

__all__: t.Tuple[str, ...] = (
    "Resource",
    "NamedResource",
)


@attrs.define(slots=True, kw_only=True)
class Resource(BaseModel):
    """Model for a resource object

    Attributes
    ----------
    url: str
        The URL of the referenced resource.
    """

    url: str = attrs.field(factory=str)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Resource":
        return cls(
            url=payload.get("url", ""),
        )


@attrs.define(slots=True, kw_only=True)
class NamedResource(BaseModel):
    """Model for a named resource object

    Attributes
    ----------
    name: str
        The name of the referenced resource.
    url: str
        The URL of the referenced resource.
    """

    name: str = attrs.field(factory=str)
    url: str = attrs.field(factory=str)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, str]) -> "NamedResource":
        return cls(name=payload.get("name", ""), url=payload.get("url", ""))



import typing as t

from ._base import BaseExtension
from .berry import Berry
from .contest import Contest
from .encounter import Encounter
from .evolution import Evolution
from .game import Game
from .item import Item
from .location import Location
from .machine import Machine
from .move import Move
from .pokemon import Pokemon

__all__: t.Tuple[str, ...] = (
    "BaseExtension",
    "Berry",
    "Contest",
    "Encounter",
    "Evolution",
    "Game",
    "Item",
    "Location",
    "Machine",
    "Move",
    "Pokemon",
)



import typing as t

from pokelance.http import Endpoint
from pokelance.models import Berry as BerryModel
from pokelance.models import BerryFirmness, BerryFlavor

from ._base import BaseExtension

if t.TYPE_CHECKING:
    from pokelance import PokeLance
    from pokelance.http import HttpClient


__all__: t.Tuple[str, ...] = (
    "setup",
    "Berry",
)


class Berry(BaseExtension):
    """Extension for berry related endpoints.

    Parameters
    ----------
    client: pokelance.http.HttpClient
        The client to use for requests.

    Attributes
    ----------
    cache: pokelance.cache.Berry
        The cache for this extension.

    """

    def __init__(self, client: "HttpClient") -> None:
        """Initializes the extension."""
        super().__init__(client)
        self.cache = self._cache.berry

    async def setup(self) -> None:
        """Sets up the extension."""
        data = await self._client.request(Endpoint.get_berry_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "berry", data["results"])
        data = await self._client.request(Endpoint.get_berry_flavor_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "berry_flavor", data["results"])
        data = await self._client.request(Endpoint.get_berry_firmness_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "berry_firmness", data["results"])

    def get_berry(self, name: t.Union[str, int]) -> t.Optional[BerryModel]:
        """Gets a berry from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the berry.

        Returns
        -------
        typing.Optional[pokelance.models.Berry]
            The berry if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the berry is invalid.

        Examples
        -------
        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> berry = client.berry.get_berry("cheri")  # None if not cached
        >>> berry.name
        'cheri'
        """
        route = Endpoint.get_berry(name)
        self._validate_resource(self.cache.berry, name, route)
        return self.cache.berry.get(route, None)

    async def fetch_berry(self, name: t.Union[str, int]) -> t.Optional[BerryModel]:
        """Fetches a berry from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the berry.

        Returns
        -------
        typing.Optional[pokelance.models.Berry]
            The berry if it exists in the API, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the berry is invalid.

        Examples
        -------
        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     berry = await client.berry.fetch_berry("cheri")
        ...     print(berry.name)
        ...     await client.close()
        >>> asyncio.run(main())
        cheri
        """
        route = Endpoint.get_berry(name)
        self._validate_resource(self.cache.berry, name, route)
        data = await self._client.request(route)
        return self.cache.berry.setdefault(route, BerryModel.from_payload(data))

    def get_berry_flavor(self, name: t.Union[str, int]) -> t.Optional[BerryFlavor]:
        """Gets a berry flavor from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the berry flavor.

        Returns
        -------
        typing.Optional[pokelance.models.BerryFlavor]
            The berry flavor if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the berry flavor is invalid.

        Examples
        -------
        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> berry_flavor = client.berry.get_berry_flavor("spicy")  # None if not cached
        >>> berry_flavor.name
        'spicy'
        """
        route = Endpoint.get_berry_flavor(name)
        self._validate_resource(self.cache.berry_flavor, name, route)
        return self.cache.berry_flavor.get(route, None)

    async def fetch_berry_flavor(self, name: t.Union[str, int]) -> t.Optional[BerryFlavor]:
        """Fetches a berry flavor from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the berry flavor.

        Returns
        -------
        typing.Optional[pokelance.models.BerryFlavor]
            The berry flavor if it exists in the API, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the berry flavor is invalid.

        Examples
        -------
        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     berry_flavor = await client.berry.fetch_berry_flavor("spicy")
        ...     print(berry_flavor.name)
        ...     await client.close()
        >>> asyncio.run(main())
        spicy
        """
        route = Endpoint.get_berry_flavor(name)
        self._validate_resource(self.cache.berry_flavor, name, route)
        data = await self._client.request(route)
        self.cache.berry_flavor[route] = BerryFlavor.from_payload(data)
        return self.cache.berry_flavor[route]

    def get_berry_firmness(self, name: t.Union[str, int]) -> t.Optional[BerryFirmness]:
        """Gets a berry firmness from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the berry firmness.

        Returns
        -------
        typing.Optional[pokelance.models.BerryFirmness]
            The berry firmness if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the berry firmness is invalid.

        Examples
        -------
        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> berry_firmness = client.berry.get_berry_firmness("very-soft")  # None if not cached
        >>> berry_firmness.name
        'very-soft'
        """
        route = Endpoint.get_berry_firmness(name)
        self._validate_resource(self.cache.berry_firmness, name, route)
        return self.cache.berry_firmness.get(route, None)

    async def fetch_berry_firmness(self, name: t.Union[str, int]) -> t.Optional[BerryFirmness]:
        """Fetches a berry firmness from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the berry firmness.

        Returns
        -------
        typing.Optional[pokelance.models.BerryFirmness]
            The berry firmness if it exists in the API, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the berry firmness is invalid.

        Examples
        -------
        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     berry_firmness = await client.berry.fetch_berry_firmness("very-soft")
        ...     print(berry_firmness.name)
        ...     await client.close()
        >>> asyncio.run(main())
        very-soft
        """
        route = Endpoint.get_berry_firmness(name)
        self._validate_resource(self.cache.berry_firmness, name, route)
        data = await self._client.request(route)
        self.cache.berry_firmness[route] = BerryFirmness.from_payload(data)
        return self.cache.berry_firmness[route]


def setup(lance: "PokeLance") -> None:
    """Sets up the berry cog."""
    lance.add_extension("berry", Berry(lance.http))



import typing as t

from pokelance.http import Endpoint
from pokelance.models import Item as ItemModel
from pokelance.models import ItemAttribute, ItemCategory, ItemFlingEffect, ItemPocket

from ._base import BaseExtension

if t.TYPE_CHECKING:
    from pokelance import PokeLance
    from pokelance.http import HttpClient


__all__: t.Tuple[str, ...] = (
    "setup",
    "Item",
)


class Item(BaseExtension):
    """Extension for item related endpoints.

    Parameters
    ----------
    client: pokelance.http.HttpClient
        The client to use for requests.

    Attributes
    ----------
    cache: pokelance.cache.Item
        The cache for this extension.
    """

    def __init__(self, client: "HttpClient") -> None:
        """Initializes the extension."""
        super().__init__(client)
        self.cache = self._cache.item

    async def setup(self) -> None:
        """Sets up the extension."""
        data = await self._client.request(Endpoint.get_item_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "item", data["results"])
        data = await self._client.request(Endpoint.get_item_attribute_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "item_attribute", data["results"])
        data = await self._client.request(Endpoint.get_item_category_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "item_category", data["results"])
        data = await self._client.request(Endpoint.get_item_fling_effect_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "item_fling_effect", data["results"])
        data = await self._client.request(Endpoint.get_item_pocket_endpoints())
        self._cache.load_documents(str(self.__class__.__name__), "item_pocket", data["results"])

    def get_item(self, name: t.Union[str, int]) -> t.Optional[ItemModel]:
        """Gets an item from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the item.

        Returns
        -------
        typing.Optional[pokelance.models.Item]
            The item if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the item is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> item = client.item.get_item("potion")
        >>> item.id
        17
        """
        route = Endpoint.get_item(name)
        self._validate_resource(self.cache.item, name, route)
        return self.cache.item.get(route, None)

    async def fetch_item(self, name: t.Union[str, int]) -> t.Optional[ItemModel]:
        """Fetches an item from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the item.

        Returns
        -------
        t.Optional[pokelance.models.Item]
            The item.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the item is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     item = await client.item.fetch_item("potion")
        ...     print(item.id)
        ...     await client.close()
        >>> asyncio.run(main())
        17
        """
        route = Endpoint.get_item(name)
        self._validate_resource(self.cache.item, name, route)
        data = await self._client.request(route)
        return self.cache.item.setdefault(route, ItemModel.from_payload(data))

    def get_item_attribute(self, name: t.Union[str, int]) -> t.Optional[ItemAttribute]:
        """Gets an item attribute from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the item attribute.

        Returns
        -------
        typing.Optional[pokelance.models.ItemAttribute]
            The item attribute if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the item attribute is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> item_attribute = client.item.get_item_attribute("holdable")
        >>> item_attribute.id
        5
        """
        route = Endpoint.get_item_attribute(name)
        self._validate_resource(self.cache.item_attribute, name, route)
        return self.cache.item_attribute.get(route, None)

    async def fetch_item_attribute(self, name: t.Union[str, int]) -> t.Optional[ItemAttribute]:
        """Fetches an item attribute from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the item attribute.

        Returns
        -------
        t.Optional[pokelance.models.ItemAttribute]
            The item attribute.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the item attribute is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     item_attribute = await client.item.fetch_item_attribute("holdable")
        ...     print(item_attribute.id)
        ...     await client.close()
        >>> asyncio.run(main())
        5
        """
        route = Endpoint.get_item_attribute(name)
        self._validate_resource(self.cache.item_attribute, name, route)
        data = await self._client.request(route)
        return self.cache.item_attribute.setdefault(route, ItemAttribute.from_payload(data))

    def get_item_category(self, name: t.Union[str, int]) -> t.Optional[ItemCategory]:
        """Gets an item category from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the item category.

        Returns
        -------
        typing.Optional[pokelance.models.ItemCategory]
            The item category if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the item category is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> item_category = client.item.get_item_category(1)
        >>> item_category.name
        'stat-boosts'
        """
        route = Endpoint.get_item_category(name)
        self._validate_resource(self.cache.item_category, name, route)
        return self.cache.item_category.get(route, None)

    async def fetch_item_category(self, name: t.Union[str, int]) -> t.Optional[ItemCategory]:
        """Fetches an item category from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the item category.

        Returns
        -------
        t.Optional[pokelance.models.ItemCategory]
            The item category.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the item category is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     item_category = await client.item.fetch_item_category(1)
        ...     print(item_category.name)
        ...     await client.close()
        >>> asyncio.run(main())
        stat-boosts
        """
        route = Endpoint.get_item_category(name)
        self._validate_resource(self.cache.item_category, name, route)
        data = await self._client.request(route)
        return self.cache.item_category.setdefault(route, ItemCategory.from_payload(data))

    def get_item_fling_effect(self, name: t.Union[str, int]) -> t.Optional[ItemFlingEffect]:
        """Gets an item fling effect from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the item fling effect.

        Returns
        -------
        typing.Optional[pokelance.models.ItemFlingEffect]
            The item fling effect if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the item fling effect is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> item_fling_effect = client.item.get_item_fling_effect(1)
        >>> item_fling_effect.name
        'badly-poison'
        """
        route = Endpoint.get_item_fling_effect(name)
        self._validate_resource(self.cache.item_fling_effect, name, route)
        return self.cache.item_fling_effect.get(route, None)

    async def fetch_item_fling_effect(self, name: t.Union[str, int]) -> t.Optional[ItemFlingEffect]:
        """Fetches an item fling effect from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the item fling effect.

        Returns
        -------
        t.Optional[pokelance.models.ItemFlingEffect]
            The item fling effect.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the item fling effect is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     item_fling_effect = await client.item.fetch_item_fling_effect(1)
        ...     print(item_fling_effect.name)
        ...     await client.close()
        >>> asyncio.run(main())
        badly-poison
        """
        route = Endpoint.get_item_fling_effect(name)
        self._validate_resource(self.cache.item_fling_effect, name, route)
        data = await self._client.request(route)
        return self.cache.item_fling_effect.setdefault(route, ItemFlingEffect.from_payload(data))

    def get_item_pocket(self, name: t.Union[str, int]) -> t.Optional[ItemPocket]:
        """Gets an item pocket from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the item pocket.

        Returns
        -------
        typing.Optional[pokelance.models.ItemPocket]
            The item pocket if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the item pocket is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> item_pocket = client.item.get_item_pocket(1)
        >>> item_pocket.name
        'misc'
        """
        route = Endpoint.get_item_pocket(name)
        self._validate_resource(self.cache.item_pocket, name, route)
        return self.cache.item_pocket.get(route, None)

    async def fetch_item_pocket(self, name: t.Union[str, int]) -> t.Optional[ItemPocket]:
        """Fetches an item pocket from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the item pocket.

        Returns
        -------
        t.Optional[pokelance.models.ItemPocket]
            The item pocket.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the item pocket is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     item_pocket = await client.item.fetch_item_pocket(1)
        ...     print(item_pocket.name)
        ...     await client.close()
        >>> asyncio.run(main())
        misc
        """
        route = Endpoint.get_item_pocket(name)
        self._validate_resource(self.cache.item_pocket, name, route)
        data = await self._client.request(route)
        return self.cache.item_pocket.setdefault(route, ItemPocket.from_payload(data))


def setup(lance: "PokeLance") -> None:
    """Sets up the item cog."""
    lance.add_extension("item", Item(lance.http))



import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import Effect, NamedResource

__all__: t.Tuple[str, ...] = (
    "AbilityEffectChange",
    "AbilityFlavorText",
    "AbilityPokemon",
    "PokemonSpeciesGender",
    "GrowthRateExperienceLevel",
    "NatureStatChange",
    "MoveBattleStylePreference",
    "NaturePokeathlonStatAffect",
    "NaturePokeathlonStatAffectSet",
    "PokemonMoveVersion",
    "PokemonStat",
    "PokemonType",
    "PokemonHeldItemVersion",
    "PokemonHeldItem",
    "PokemonSprite",
    "PokemonAbility",
    "PokemonMove",
    "PokemonTypePast",
    "PokemonFormType",
    "PokemonFormSprites",
    "AwesomeName",
    "PokemonSpeciesVariety",
    "PokemonSpeciesDexEntry",
    "PalParkEncounterArea",
    "Genus",
    "MoveStatAffectSets",
    "NatureStatAffectSets",
    "MoveStatEffect",
    "TypeRelations",
    "TypePokemon",
    "TypeRelationsPast",
)


@attrs.define(slots=True, kw_only=True)
class AbilityEffectChange(BaseModel):
    """An ability effect change resource.

    Attributes
    ----------
    effect_entries: t.List[Effect]
        The previous effect of this ability listed in different languages.
    version_group: NamedResource
        The version group in which the previous effect of this ability originated.
    """

    effect_entries: t.List[Effect] = attrs.field(factory=list)
    version_group: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "AbilityEffectChange":
        return cls(
            effect_entries=[Effect.from_payload(i) for i in payload.get("effect_entries", [])],
            version_group=NamedResource.from_payload(payload.get("version_group", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class AbilityFlavorText(BaseModel):
    """An ability flavor text resource.

    Attributes
    ----------
    flavor_text: str
        The localized flavor text for an api resource in a specific language.
    language: NamedResource
        The language this name is in.
    version_group: NamedResource
        The version group that uses this flavor text.
    """

    flavor_text: str = attrs.field(factory=str)
    language: NamedResource = attrs.field(factory=NamedResource)
    version_group: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "AbilityFlavorText":
        return cls(
            flavor_text=payload.get("flavor_text", ""),
            language=NamedResource.from_payload(payload.get("language", {}) or {}),
            version_group=NamedResource.from_payload(payload.get("version_group", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class AbilityPokemon(BaseModel):
    """An ability pokemon resource.

    Attributes
    ----------
    is_hidden: bool
        Whether or not this a hidden ability for the referenced pokemon.
    slot: int
        The slot this ability occupies in this pokemon species.
    pokemon: NamedResource
        The pokemon this ability could belong to.
    """

    is_hidden: bool = attrs.field(factory=bool)
    slot: int = attrs.field(factory=int)
    pokemon: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "AbilityPokemon":
        return cls(
            is_hidden=payload.get("is_hidden", False),
            slot=payload.get("slot", 0),
            pokemon=NamedResource.from_payload(payload.get("pokemon", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class PokemonSpeciesGender(BaseModel):
    """A pokemon species gender resource.

    Attributes
    ----------
    rate: int
        The chance of a particular gender in this pokemon species.
    pokemon_species: NamedResource
        The pokemon species here.
    """

    rate: int = attrs.field(factory=int)
    pokemon_species: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonSpeciesGender":
        return cls(
            rate=payload.get("rate", 0),
            pokemon_species=NamedResource.from_payload(payload.get("pokemon_species", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class GrowthRateExperienceLevel(BaseModel):
    """A growth rate experience level resource.

    Attributes
    ----------
    level: int
        The level gained.
    experience: int
        The amount of experience required to reach the referenced level.
    """

    level: int = attrs.field(factory=int)
    experience: int = attrs.field(factory=int)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "GrowthRateExperienceLevel":
        return cls(
            level=payload.get("level", 0),
            experience=payload.get("experience", 0),
        )


@attrs.define(slots=True, kw_only=True)
class NatureStatChange(BaseModel):
    """A nature stat change resource.

    Attributes
    ----------
    max_change: int
        The maximum amount of change to the referenced stat.
    pokeathlon_stat: NamedResource
        The stat being affected.
    """

    max_change: int = attrs.field(factory=int)
    pokeathlon_stat: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "NatureStatChange":
        return cls(
            max_change=payload.get("max_change", 0),
            pokeathlon_stat=NamedResource.from_payload(payload.get("pokeathlon_stat", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class MoveBattleStylePreference(BaseModel):
    """A move battle style preference resource.

    Attributes
    ----------
    low_hp_preference: int
        Chance of using the move, in percent, if HP is under one half.
    high_hp_preference: int
        Chance of using the move, in percent, if HP is over one half.
    move_battle_style: NamedResource
        The move battle style.
    """

    low_hp_preference: int = attrs.field(factory=int)
    high_hp_preference: int = attrs.field(factory=int)
    move_battle_style: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "MoveBattleStylePreference":
        return cls(
            low_hp_preference=payload.get("low_hp_preference", 0),
            high_hp_preference=payload.get("high_hp_preference", 0),
            move_battle_style=NamedResource.from_payload(payload.get("battle_style", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class NaturePokeathlonStatAffect(BaseModel):
    """A nature pokeathlon stat affect resource.

    Attributes
    ----------
    max_change: int
        The maximum amount of change to the referenced stat.
    nature: NamedResource
        The nature causing the change.
    """

    max_change: int = attrs.field(factory=int)
    nature: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "NaturePokeathlonStatAffect":
        return cls(
            max_change=payload.get("max_change", 0),
            nature=NamedResource.from_payload(payload.get("nature", {})),
        )


@attrs.define(slots=True, kw_only=True)
class NaturePokeathlonStatAffectSet(BaseModel):
    """A nature pokeathlon stat affect set resource.

    Attributes
    ----------
    increase: t.List[NaturePokeathlonStatAffect]
        A list of natures and how they change the referenced pokeathlon stat.
    decrease: t.List[NaturePokeathlonStatAffect]
        A list of natures and how they change the referenced pokeathlon stat.
    """

    increase: t.List[NaturePokeathlonStatAffect] = attrs.field(factory=list)
    decrease: t.List[NaturePokeathlonStatAffect] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "NaturePokeathlonStatAffectSet":
        return cls(
            increase=[NaturePokeathlonStatAffect.from_payload(i) for i in payload.get("increase", [])],
            decrease=[NaturePokeathlonStatAffect.from_payload(i) for i in payload.get("decrease", [])],
        )


@attrs.define(slots=True, kw_only=True)
class PokemonAbility(BaseModel):
    """A pokemon ability resource.

    Attributes
    ----------
    is_hidden: bool
        Whether or not this is a hidden ability.
    slot: int
        The slot this ability occupies in this pokemon species.
    ability: NamedResource
        The ability the pokemon may have.
    """

    is_hidden: bool = attrs.field(factory=bool)
    slot: int = attrs.field(factory=int)
    ability: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonAbility":
        return cls(
            is_hidden=payload.get("is_hidden", False),
            slot=payload.get("slot", 0),
            ability=NamedResource.from_payload(payload.get("ability", {})),
        )


@attrs.define(slots=True, kw_only=True)
class PokemonType(BaseModel):
    """A pokemon type resource.

    Attributes
    ----------
    slot: int
        The order the types are listed in.
    type: NamedResource
        The type the referenced pokemon has.
    """

    slot: int = attrs.field(factory=int)
    type: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonType":
        return cls(
            slot=payload.get("slot", 0),
            type=NamedResource.from_payload(payload.get("type", {})),
        )


@attrs.define(slots=True, kw_only=True)
class PokemonFormType(BaseModel):
    """A pokemon form type resource.

    Attributes
    ----------
    slot: int
        The order the types are listed in.
    type: NamedResource
        The type the referenced pokemon has.
    """

    slot: int = attrs.field(factory=int)
    type: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonFormType":
        return cls(
            slot=payload.get("slot", 0),
            type=NamedResource.from_payload(payload.get("type", {})),
        )


@attrs.define(slots=True, kw_only=True)
class PokemonTypePast(BaseModel):
    """A pokemon type past resource.

    Attributes
    ----------
    generation: NamedResource
        The generation this type was introduced in.
    types: t.List[NamedResource]
        The name of the type.
    """

    generation: NamedResource = attrs.field(factory=NamedResource)
    types: t.List[PokemonType] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonTypePast":
        return cls(
            generation=NamedResource.from_payload(payload.get("generation", {})),
            types=[PokemonType.from_payload(i) for i in payload.get("types", [])],
        )


@attrs.define(slots=True, kw_only=True)
class PokemonHeldItemVersion(BaseModel):
    """A pokemon held item version resource.

    Attributes
    ----------
    rarity: int
        How often this item is held.
    version: NamedResource
        The version this item is held in by the pokemon.
    """

    version: NamedResource = attrs.field(factory=NamedResource)
    rarity: int = attrs.field(factory=int)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonHeldItemVersion":
        return cls(
            version=NamedResource.from_payload(payload.get("version", {})),
            rarity=payload.get("rarity", 0),
        )


@attrs.define(slots=True, kw_only=True)
class PokemonHeldItem(BaseModel):
    """A pokemon held item resource.

    Attributes
    ----------
    item: NamedResource
        The item the referenced pokemon holds.
    version_details: t.List[PokemonHeldItemVersion]
        The details for the version that this item is held in by the pokemon.
    """

    item: NamedResource = attrs.field(factory=NamedResource)
    version_details: t.List[PokemonHeldItemVersion] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonHeldItem":
        return cls(
            item=NamedResource.from_payload(payload.get("item", {})),
            version_details=[PokemonHeldItemVersion.from_payload(i) for i in payload.get("version_details", [])],
        )


@attrs.define(slots=True, kw_only=True)
class PokemonMoveVersion(BaseModel):
    """A pokemon move version resource.

    Attributes
    ----------
    move_learn_method: NamedResource
        The method by which the move is learned.
    version_group: NamedResource
        The version group in which the move is learned.
    level_learned_at: int
        The minimum level to learn the move.
    """

    move_learn_method: NamedResource = attrs.field(factory=NamedResource)
    version_group: NamedResource = attrs.field(factory=NamedResource)
    level_learned_at: int = attrs.field(factory=int)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonMoveVersion":
        return cls(
            move_learn_method=NamedResource.from_payload(payload.get("move_learn_method", {})),
            version_group=NamedResource.from_payload(payload.get("version_group", {})),
            level_learned_at=payload.get("level_learned_at", 0),
        )


@attrs.define(slots=True, kw_only=True)
class PokemonMove(BaseModel):
    """A pokemon move resource.

    Attributes
    ----------
    move: NamedResource
        The move the referenced pokemon can learn.
    version_group_details: t.List[PokemonMoveVersion]
        The details for the version group that this move can be learned in.
    """

    move: NamedResource = attrs.field(factory=NamedResource)
    version_group_details: t.List[PokemonMoveVersion] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonMove":
        return cls(
            move=NamedResource.from_payload(payload.get("move", {})),
            version_group_details=[
                PokemonMoveVersion.from_payload(i) for i in payload.get("version_group_details", [])
            ],
        )


@attrs.define(slots=True, kw_only=True)
class PokemonStat(BaseModel):
    """A pokemon stat resource.

    Attributes
    ----------
    stat: NamedResource
        The stat the referenced pokemon has.
    effort: int
        The effort points (EV) the referenced pokemon has in the stat.
    base_stat: int
        The base value of the stat.
    """

    stat: NamedResource = attrs.field(factory=NamedResource)
    effort: int = attrs.field(factory=int)
    base_stat: int = attrs.field(factory=int)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonStat":
        return cls(
            stat=NamedResource.from_payload(payload.get("stat", {})),
            effort=payload.get("effort", 0),
            base_stat=payload.get("base_stat", 0),
        )


@attrs.define(slots=True, kw_only=True)
class PokemonSprite(BaseModel):
    """A pokemon sprite resource.

    Attributes
    ----------
    front_default: str
        The default depiction of this pokemon from the front in battle.
    front_shiny: str
        The shiny depiction of this pokemon from the front in battle.
    front_female: str
        The default depiction of female gender of this pokemon from the front in battle.
    front_shiny_female: str
        The shiny depiction of female gender of this pokemon from the front in battle.
    back_default: str
        The default depiction of this pokemon from the back in battle.
    back_shiny: str
        The shiny depiction of this pokemon from the back in battle.
    back_female: str
        The default depiction of female gender of this pokemon from the back in battle.
    back_shiny_female: str
        The shiny depiction of female gender of this pokemon from the back in battle.
    """

    front_default: str = attrs.field(factory=str)
    front_shiny: str = attrs.field(factory=str)
    front_female: str = attrs.field(factory=str)
    front_shiny_female: str = attrs.field(factory=str)
    back_default: str = attrs.field(factory=str)
    back_shiny: str = attrs.field(factory=str)
    back_female: str = attrs.field(factory=str)
    back_shiny_female: str = attrs.field(factory=str)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonSprite":
        return cls(
            front_default=payload.get("front_default", ""),
            front_shiny=payload.get("front_shiny", ""),
            front_female=payload.get("front_female", ""),
            front_shiny_female=payload.get("front_shiny_female", ""),
            back_default=payload.get("back_default", ""),
            back_shiny=payload.get("back_shiny", ""),
            back_female=payload.get("back_female", ""),
            back_shiny_female=payload.get("back_shiny_female", ""),
        )


@attrs.define(slots=True, kw_only=True)
class PokemonFormSprites(BaseModel):
    """A pokemon form sprites resource.

    Attributes
    ----------
    front_default: str
        The default depiction of this pokemon form from the front in battle.
    front_shiny: str
        The shiny depiction of this pokemon form from the front in battle.
    back_default: str
        The default depiction of this pokemon form from the back in battle.
    back_shiny: str
        The shiny depiction of this pokemon form from the back in battle.
    """

    front_default: str = attrs.field(factory=str)
    front_shiny: str = attrs.field(factory=str)
    back_default: str = attrs.field(factory=str)
    back_shiny: str = attrs.field(factory=str)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonFormSprites":
        return cls(
            front_default=payload.get("front_default", ""),
            front_shiny=payload.get("front_shiny", ""),
            back_default=payload.get("back_default", ""),
            back_shiny=payload.get("back_shiny", ""),
        )


@attrs.define(slots=True, kw_only=True)
class AwesomeName(BaseModel):
    """An awesome name resource.

    Attributes
    ----------
    awesome_name: str
        The localized "scientific" name for an API resource in a specific language.
    language: NamedResource
        The language this "scientific" name is in.
    """

    awesome_name: str = attrs.field(factory=str)
    language: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "AwesomeName":
        return cls(
            awesome_name=payload.get("awesome_name", ""),
            language=NamedResource.from_payload(payload.get("language", {})),
        )


@attrs.define(slots=True, kw_only=True)
class Genus(BaseModel):
    """A genus resource.

    Attributes
    ----------
    genus: str
        The localized genus for the referenced type in the specified language.
    language: NamedResource
        The language this genus is in.
    """

    genus: str = attrs.field(factory=str)
    language: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Genus":
        return cls(
            genus=payload.get("genus", ""),
            language=NamedResource.from_payload(payload.get("language", {})),
        )


@attrs.define(slots=True, kw_only=True)
class PokemonSpeciesDexEntry(BaseModel):
    """A pokemon species dex entry resource.

    Attributes
    ----------
    entry_number: int
        The index number within the Pokédex.
    pokedex: NamedResource
        The Pokédex the referenced Pokémon species can be found in.
    """

    entry_number: int = attrs.field(factory=int)
    pokedex: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonSpeciesDexEntry":
        return cls(
            entry_number=payload.get("entry_number", 0),
            pokedex=NamedResource.from_payload(payload.get("pokedex", {})),
        )


@attrs.define(slots=True, kw_only=True)
class PalParkEncounterArea(BaseModel):
    """A pal park encounter area resource.

    Attributes
    ----------
    base_score: int
        The base score given to the player when the referenced Pokémon is caught during a pal park run.
    rate: int
        The base rate for encountering the referenced Pokémon in this pal park area.
    area: NamedResource
        The pal park area where this encounter happens.
    """

    base_score: int = attrs.field(factory=int)
    rate: int = attrs.field(factory=int)
    area: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PalParkEncounterArea":
        return cls(
            base_score=payload.get("base_score", 0),
            rate=payload.get("rate", 0),
            area=NamedResource.from_payload(payload.get("area", {})),
        )


@attrs.define(slots=True, kw_only=True)
class PokemonSpeciesVariety(BaseModel):
    """A pokemon species variety resource.

    Attributes
    ----------
    is_default: bool
        Whether this variety is the default variety.
    pokemon: NamedResource
        The Pokémon variety.
    """

    is_default: bool = attrs.field(factory=bool)
    pokemon: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonSpeciesVariety":
        return cls(
            is_default=payload.get("is_default", False),
            pokemon=NamedResource.from_payload(payload.get("pokemon", {})),
        )


@attrs.define(slots=True, kw_only=True)
class MoveStatEffect(BaseModel):
    """A move stat effect resource.

    Attributes
    ----------
    change: int
        The maximum amount of change to the referenced stat.
    stat: NamedResource
        The stat being affected.
    """

    change: int = attrs.field(factory=int)
    stat: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "MoveStatEffect":
        return cls(
            change=payload.get("change", 0),
            stat=NamedResource.from_payload(payload.get("stat", {})),
        )


@attrs.define(slots=True, kw_only=True)
class MoveStatAffectSets(BaseModel):
    """A move stat affect sets resource.

    Attributes
    ----------
    increase: typing.List[MoveStatAffect]
        A list of move stat affects.
    decrease: typing.List[MoveStatAffect]
        A list of move stat affects.
    """

    increase: t.List[MoveStatEffect] = attrs.field(factory=list)
    decrease: t.List[MoveStatEffect] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "MoveStatAffectSets":
        return cls(
            increase=[MoveStatEffect.from_payload(i) for i in payload.get("increase", [])],
            decrease=[MoveStatEffect.from_payload(i) for i in payload.get("decrease", [])],
        )


@attrs.define(slots=True, kw_only=True)
class NatureStatAffectSets(BaseModel):
    """A nature stat affect sets resource.

    Attributes
    ----------
    increase: typing.List[NatureStatAffect]
        A list of nature stat affects.
    decrease: typing.List[NatureStatAffect]
        A list of nature stat affects.
    """

    increase: t.List[NamedResource] = attrs.field(factory=list)
    decrease: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "NatureStatAffectSets":
        return cls(
            increase=[NamedResource.from_payload(i) for i in payload.get("increase", [])],
            decrease=[NamedResource.from_payload(i) for i in payload.get("decrease", [])],
        )


@attrs.define(slots=True, kw_only=True)
class TypePokemon(BaseModel):
    """A type pokemon resource.

    Attributes
    ----------
    slot: int
        The order the Pokémon's types are listed in.
    pokemon: NamedResource
        The Pokémon that has the referenced type.
    """

    slot: int = attrs.field(factory=int)
    pokemon: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "TypePokemon":
        return cls(
            slot=payload.get("slot", 0),
            pokemon=NamedResource.from_payload(payload.get("pokemon", {})),
        )


@attrs.define(slots=True, kw_only=True)
class TypeRelations(BaseModel):
    """A type relations resource.

    Attributes
    ----------
    no_damage_to: typing.List[NamedResource]
        A list of types this type has no damage to.
    half_damage_to: typing.List[NamedResource]
        A list of types this type is half damage to.
    double_damage_to: typing.List[NamedResource]
        A list of types this type is double damage to.
    no_damage_from: typing.List[NamedResource]
        A list of types that have no damage to this type.
    half_damage_from: typing.List[NamedResource]
        A list of types that have half damage to this type.
    double_damage_from: typing.List[NamedResource]
        A list of types that have double damage to this type.
    """

    no_damage_to: t.List[NamedResource] = attrs.field(factory=list)
    half_damage_to: t.List[NamedResource] = attrs.field(factory=list)
    double_damage_to: t.List[NamedResource] = attrs.field(factory=list)
    no_damage_from: t.List[NamedResource] = attrs.field(factory=list)
    half_damage_from: t.List[NamedResource] = attrs.field(factory=list)
    double_damage_from: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "TypeRelations":
        return cls(
            no_damage_to=[NamedResource.from_payload(i) for i in payload.get("no_damage_to", [])],
            half_damage_to=[NamedResource.from_payload(i) for i in payload.get("half_damage_to", [])],
            double_damage_to=[NamedResource.from_payload(i) for i in payload.get("double_damage_to", [])],
            no_damage_from=[NamedResource.from_payload(i) for i in payload.get("no_damage_from", [])],
            half_damage_from=[NamedResource.from_payload(i) for i in payload.get("half_damage_from", [])],
            double_damage_from=[NamedResource.from_payload(i) for i in payload.get("double_damage_from", [])],
        )


@attrs.define(slots=True, kw_only=True)
class TypeRelationsPast(BaseModel):
    """A type relations past resource.

    Attributes
    ----------
    generation: NamedResource
        The generation of this type relation.
    damage_relations: TypeRelations
        The type relations for this generation.
    """

    generation: NamedResource = attrs.field(factory=NamedResource)
    damage_relations: TypeRelations = attrs.field(factory=TypeRelations)
